% 
% The driver function to test spectral regularization on
% orthographic structure and motion.
%
% In the case of no noise one would expect the method to achieve
% similar performance as bundle adjustment. I have yet to be able
% to show this.
%
% Some of the code to display reprojections may be commented out
% (look for XXX)
% TODO: make it work better, analytic bundle adjustment gradients.
%
% (c) Neil Birkbeck 2010
function [avgdist, cubedist] = orthocam(noise_level, pvis, nextra, bundle_adjust)

if nargin < 1
  noise_level = 0.00;
end

if nargin < 2
  pvis = 0.1;
end

if nargin < 3,
  nextra = 10
end

if nargin < 4,
  bundle_adjust = false;
end

% Generate a cube.
X=[-1,-1,-1;
   1,-1,-1;
   1, 1,-1;
   -1, 1,-1;
   -1,-1, 1;
   1,-1, 1;
   1, 1, 1;
   -1, 1, 1]';

% Use these lines to index the cube
lines=[1,2,3,4,1,5,6,7,8,5,6,2,6,7,3,7,8,4];

% Add some extra points, just to be sure that we have enough
% visible points.
X = [X, rand(3, nextra)*2 - 1];
X(4, :) = 1;

P = generate_cameras(10);

obs = P*X;
obs = obs + (rand(size(obs))*2 - 1)*noise_level;

% Generate a random visibility matrix such that the 
% x,y coords for a point are either vis or invis
V = rand(size(obs,1)/2,  size(obs, 2)) > pvis;
V = V';
V = V(:)';
V = reshape([V;V], size(obs));


fprintf('Prob vis: %f\n', sum(V(:))/prod(size(V)));

%[P2,X2] = factorize_orig(obs.*V, double(V), bundle_adjust);
[P2,X2] = factorize(obs, V, bundle_adjust);
%[P2,X2] = orthofact(obs, mean(obs, 2) + (rand(size(obs, 1), 1) - 0.5)*0.1);


figure(1);
plot3(X(1,lines), X(2,lines), X(3,lines), 'r');
hold on;
plot3(X2(1,lines), X2(2,lines), X2(3,lines));
hold off;

axis equal;

figure(2);
diff = obs - P2*X2;
avgdist = sqrt(dot(diff(:), diff(:))/(size(diff, 1)*size(diff, 2)/2));
fprintf('Average distance: %f\n', avgdist);
% XXX: uncomment to show reprojection errors.
% show_projections(obs, V, P2*X2, lines);


% x = [0,0,0,0,0.1,0.1,1];
% f=@(x)(evaluate_metric(x, X(:,1:8), X2(:,1:8)))
% x = lsqnonlin(f, x)
T = estimate_affine(X, X2); %extract_tform(x)
X3 = T*X2;

figure(1);
plot3(X(1,lines), X(2,lines), X(3,lines), 'r');
hold on;
plot3(X3(1,lines), X3(2,lines), X3(3,lines));
title('Aligned cubes.');
hold off;

cubedist = norm(X3(:,1:8) - X(:,1:8), 'fro')/8;


function show_proj_error(X, V, Z, lambda, message)
S = svd(Z);
[etot, edata, ereg] = proj_error(X, V, Z,lambda);
fprintf('%s: %f = data(%f) + reg(%f), (lambda = %f)\n', message, ...
        etot, edata, ereg, lambda);
S(1:5)'


function [etot, edata, ereg] = proj_error(X, V, Z, lambda)
edata = 0.5*norm((X - Z).*V, 'fro');
ereg = lambda * sum(svd(Z));
etot = edata + ereg;

function [Zs, lambda] = find_first_lambda_with_rank(data, V, lambda, therank,thresh, init)

while 1 
  Zs = spectreg(data, V, [lambda], thresh, init);
  if (rank(Zs{1})<=therank),
    break
  end
  lambda = lambda + 0.00025;
  init=Zs{1};
end



% Return an affine transformation that aligns the two point sets.
function T = estimate_affine(X1, X2)
n = size(X2, 2);
A = reshape([X2(1:4, :); zeros(12, n); X2(1:4, :); zeros(12, n); X2(1:4, :)],12,n*3)';
b = X1(1:3,:);
b = b(:);
T = A\b;
T = reshape(T, 4, 3)';
T(4,:) = [0,0,0,1];




% Extract a metric transformation from a packed vector
function T = extract_metric(x)
T = [rodrigues(x(1:3)) x(4:6)'];
T(4,:) = [0 0 0 1];
T =  T * diag([x(7), x(7), x(7), 1]);




% Evaluate the results of a metric transform
function f = evaluate_metric(x, X1, X2)
T = extract_metric(x);
f = X1 - T * X2;
f = f(1:3,:);
f = f(:);



% Display the projections.
function show_projections(obs1, vis, obs2, lines)
for i=1:2:size(obs1, 1),
  o1 = obs1(i:(i+1),:);
  o2 = obs2(i:(i+1),:);
    
  plot(o1(1,lines), o1(2,lines), 'r-');
  hold on;
  
  vx = find(vis(i, :));
  vy = find(vis(i+1, :));
  plot(o1(1, :), o1(2, :), 'r.');
  plot(o1(1, vx), o1(2, vx), 'r*');
  
  
  plot(o2(1,lines), o2(2,lines), 'g:o');
  hold off;
  drawnow;
  waitforbuttonpress;
end


% Perform the factorization with spectral filling.
% Notice how the spectreg is bootstrapped with the previous
% solution, and also how the centroid of the points is computed
% for two rank updates.  This appeared to be necessary.
%
% Finally, it does something like the spectral regularization but
% uses the orthographic camera factorization to do the projection.
%
% See the original function factorize_orig
function [P,X] = factorize(obs, V, bundle_adjust)
lambda=0.0001;
init = obs.*V;

% Find a good first rank 4 approximation
[Zs, lambda] = find_first_lambda_with_rank(obs.*V, V, lambda, 4, 1e-6, init);

show_proj_error(obs, V, obs, lambda, 'original (rank 4):');

% Refine the rank 4 solution a bit.
Zs = spectreg(obs .* V, V, [lambda], 1e-25, Zs{1});
show_proj_error(obs, V, Zs{1}, lambda, 'spect (rank 4, found): ');

% Check from different starting positions (e.g, starting from the
% true obs).
Zobs = spectreg(obs .* V, V, [lambda], 1e-25, obs);
show_proj_error(obs, V, Zobs{1}, lambda, 'spect (rank 4, start truth): ');


% Calculate the mean, going to start rank 3 approximation.
mn = mean(Zs{1}, 2);

lambda = max(lambda - 0.0002, 0.0001);
mn_obs = obs - repmat(mn, [1, size(obs, 2)]);
[Zs, lambda] = find_first_lambda_with_rank(mn_obs.*V, V, lambda, 3, 1e-6,Zs{1});

% Show some residuals.
show_proj_error(mn_obs, V, obs - repmat(mean(obs, 2), [1, size(obs, 2)]), lambda, 'original (rank 3):');

Zs = spectreg(mn_obs .* V, V, [lambda], 1e-25, Zs{1});
show_proj_error(mn_obs, V, Zs{1}, lambda, 'spect (rank 3):');

Zs = spectreg(mn_obs .* V, V, [lambda], 1e-25, Zs{1}, 1);
show_proj_error(mn_obs, V, Zs{1}, lambda, 'spect (rank 3, hard):');

% Add back the mean, and use Zs{1} as a  filled in estimate.
Zs{1} = Zs{1} + repmat(mn, [1, size(obs, 2)])
Zs{1} = Zs{1}.*(~V) + obs.*V;

% Refine the solution.
last = Zs{1};
epsilon = 1e-9;

maxits=1
for i=1:maxits,
  [P,X] = orthofact(Zs{1}, [], false);
  
  Zs{1} = P*X;
  
  if mod(i, 100)==0,
    fprintf('res: %f, %f \n', norm((Zs{1} - obs).*(~V), 'fro'),...
            norm((Zs{1} - obs).*(V), 'fro'),...
            norm((Zs{1} - obs), 'fro'));
  end
  
  % Only keep the unknown stuff.
  Zs{1} = (P*X).*(~V) + obs.*V;
  
  err = ((norm(Zs{1} - last, 'fro')) / (norm(last, 'fro')))^2;
  if err < epsilon,
    break;
  end  
end

% Do bundle adjustment.
if bundle_adjust
  Y = pack_cameras_and_points(P, X);
  f = @(x)(bundle_func([size(P, 1)/2, size(X, 2)], x, V, obs));
  Y = fminunc(f, Y)
  [P, X] = extract_cameras_and_points([size(P, 1)/2, size(X, 2)], Y) 
end
return;




% Perform the factorization with spectral filling.
% Notice how the spectreg is bootstrapped with the previous
% solution, and also how the centroid of the points is computed
% for several rank updates.  This appeared to be necessary.
%
% TODO: find the "best" lambda that satisfies the rank (e.g.,
% lowest residual in terms of the matrix elements).
function [P,X] = factorize_orig(obs, Vis, bundle_adjust)

if nargin < 2
  Vis = ones(size(obs));
  Vis
end

%mn = mean(obs, 2); 
mn = sum(obs.*Vis , 2)./sum(Vis, 2);
mn_obs = obs - repmat(mn, [1,size(obs,2)]);
mn_obs = mn_obs .* Vis; % Just zero the stuff we don't know.

% Try to reduce the rank of the matrix slowly, hoping to 
% get a more accurate estimate of the mean and the unknown entries.
rank_break=repmat(4:-1:3,[100,1]);
rank_break=rank_break(:);
rank_break = [8,7,6,5,4,3];

last_lambda = 0.000001;

for i=1:length(rank_break),
  % Find a value of lambda that gives us the desired rank.
  lambda = last_lambda;
  
  % Evaluate the solution until we find a matching rank.
  init = mn_obs;
  while (1)
    Z = spectreg(mn_obs, Vis, [lambda], 1e-6, init);
    init = Z{1};
    
    %'Error:'
    %norm(Z{1}.*Vis - mn_obs.*Vis)
    %fprintf('%d\n', rank(Z{1}));
    
    if (rank(Z{1}) <= rank_break(i))
      % Z = spectreg(mn_obs, Vis, [lambda], 1e-20);
      break;
    end
    
    last_lambda = lambda;
    lambda = lambda + 0.001;    
  end
    
  mn = mean(Z{1} + repmat(mn, [1,size(obs,2)]), 2);
  mn_obs = obs - repmat(mn, [1,size(obs,2)]);
  mn_obs = mn_obs.*Vis + Z{1}.*(~Vis);
  fprintf('Mean difference(%d): %f\n', rank_break(i), norm(mn - mean(obs, 2)));
end

'Error on observations:'
norm(Z{1}.*Vis - mn_obs.*Vis)

% Orthofact assumes that the coordinates are not normalized yet,
% So give it to it that way
[P, X] = orthofact(mn_obs + repmat(mn, [1, size(obs, 2)]), mn);

% Do bundle adjustment.
if bundle_adjust
  Y = pack_cameras_and_points(P, X);
  f = @(x)(bundle_func([size(P, 1)/2, size(X, 2)], x, Vis, obs));
  Y = fminunc(f, Y)
  [P, X] = extract_cameras_and_points([size(P, 1)/2, size(X, 2)], Y) 
end
return;


% Generate a sequence of n cameras
function P = generate_cameras(n)
P = zeros(2*n, 4);
for i=1:n,
  p = trans(rand(1,1)-0.5, rand(1,1)-0.5, rand(1,1)-0.5)*rotx(rand(1,1)*0.9 ...
                                                    - 0.45)*roty(2*pi * i/n);
  P((2*(i-1)+1):2*i, :) = p(1:2,:);
end
return;




% Pack the cameras and points for bundle adjustment.
function Y = pack_cameras_and_points(P, X)
Y = [];
ncams=size(P, 1)/2;
for i=1:ncams,
  i1 = 2*i-1;
  i2 = 2*i;
  r = rodrigues([P(i1, 1:3);
                 P(i2, 1:3);
                 cross(P(i1, 1:3), P(i2, 1:3))]);
  Y = [Y; r(1:3); P(i1, 4); P(i2, 4)];
end
x = X(1:3, :);
Y = [Y; x(:)];
return;


% Function for unpacking the cameras and points in the bundle
% adjustment.
function [P, X, Pindy] = extract_cameras_and_points(sz, Y) 
ncams = sz(1);
npoints = sz(2);
C = reshape(Y(1:5*ncams), [5, ncams]);
X = reshape(Y((5*ncams + 1):end), [3, npoints]);
X(4,:) = 1;
Pindy =cell(ncams);
P = [];

for i=1:ncams, 
  R = rodrigues(C(1:3, i));
  Pindy{i} = [R(1,:), C(4, i);
              R(2,:), C(5, i)];
  P = [P; Pindy{i}];
end
return;


% Bundle adjustment cost function
function F = bundle_func(sz, Y, V, obs)
[P, X] = extract_cameras_and_points(sz, Y);

ncams = sz(1);
npoints = sz(2);
r = (obs - P * X).*V;
F = dot(r(:), r(:));
return
